"""Tests for yt_dbl.schemas — data models and pipeline state."""

import pytest

from yt_dbl.schemas import (
    STEP_DIRS,
    STEP_ORDER,
    PipelineState,
    Segment,
    Speaker,
    StepName,
    StepResult,
    StepStatus,
    VideoMeta,
    Word,
)


class TestWord:
    def test_creation(self) -> None:
        w = Word(text="hello", start=0.0, end=0.5)
        assert w.text == "hello"
        assert w.confidence == 1.0

    def test_custom_confidence(self) -> None:
        w = Word(text="um", start=1.0, end=1.2, confidence=0.42)
        assert w.confidence == pytest.approx(0.42)

    def test_zero_length_word(self) -> None:
        w = Word(text="", start=0.0, end=0.0)
        assert w.text == ""
        assert w.end - w.start == 0.0


class TestSegment:
    def test_duration(self) -> None:
        seg = Segment(id=0, text="test", start=1.0, end=3.5)
        assert seg.duration == 2.5

    def test_defaults(self) -> None:
        seg = Segment(id=0, text="test", start=0.0, end=1.0)
        assert seg.speaker == "SPEAKER_00"
        assert seg.words == []
        assert seg.translated_text == ""

    def test_zero_duration(self) -> None:
        seg = Segment(id=0, text="silence", start=5.0, end=5.0)
        assert seg.duration == 0.0

    def test_with_words(self) -> None:
        words = [
            Word(text="hello", start=0.0, end=0.3),
            Word(text="world", start=0.4, end=0.8),
        ]
        seg = Segment(id=1, text="hello world", start=0.0, end=0.8, words=words)
        assert len(seg.words) == 2
        assert seg.words[0].text == "hello"

    def test_translated_text_field(self) -> None:
        seg = Segment(id=0, text="Hello", start=0.0, end=1.0, translated_text="Привет")
        assert seg.translated_text == "Привет"


class TestSpeaker:
    def test_defaults(self) -> None:
        sp = Speaker(id="SPEAKER_01")
        assert sp.reference_start == 0.0
        assert sp.reference_end == 0.0
        assert sp.reference_path == ""
        assert sp.total_duration == 0.0

    def test_with_reference(self) -> None:
        sp = Speaker(
            id="SPEAKER_00",
            reference_start=10.0,
            reference_end=17.0,
            reference_path="ref.wav",
            total_duration=120.5,
        )
        assert sp.reference_end - sp.reference_start == 7.0
        assert sp.reference_path == "ref.wav"


class TestVideoMeta:
    def test_minimal(self) -> None:
        m = VideoMeta(video_id="abc")
        assert m.title == ""
        assert m.duration == 0.0

    def test_full(self) -> None:
        m = VideoMeta(
            video_id="abc",
            title="Test",
            channel="Ch",
            duration=300.0,
            url="https://youtube.com/watch?v=abc",
        )
        assert m.channel == "Ch"
        assert m.duration == 300.0


class TestStepResult:
    def test_defaults(self) -> None:
        r = StepResult(step=StepName.DOWNLOAD)
        assert r.status == StepStatus.PENDING
        assert r.started_at == ""
        assert r.finished_at == ""
        assert r.duration_sec == 0.0
        assert r.error == ""
        assert r.outputs == {}

    def test_failed_step(self) -> None:
        r = StepResult(
            step=StepName.TRANSCRIBE,
            status=StepStatus.FAILED,
            error="Model OOM",
        )
        assert r.status == StepStatus.FAILED
        assert "OOM" in r.error


class TestStepEnums:
    def test_step_order_length(self) -> None:
        assert len(STEP_ORDER) == 6

    def test_step_dirs_keys_match_order(self) -> None:
        assert list(STEP_DIRS.keys()) == STEP_ORDER

    def test_step_dirs_prefixes(self) -> None:
        for i, (_name, dirname) in enumerate(STEP_DIRS.items(), start=1):
            assert dirname.startswith(f"{i:02d}_")

    def test_status_values(self) -> None:
        assert StepStatus.PENDING.value == "pending"
        assert StepStatus.RUNNING.value == "running"
        assert StepStatus.COMPLETED.value == "completed"
        assert StepStatus.FAILED.value == "failed"


class TestPipelineState:
    def test_next_step_fresh(self) -> None:
        state = PipelineState(video_id="abc123")
        assert state.next_step == StepName.DOWNLOAD

    def test_next_step_after_download(self) -> None:
        state = PipelineState(video_id="abc123")
        step = state.get_step(StepName.DOWNLOAD)
        step.status = StepStatus.COMPLETED
        assert state.next_step == StepName.SEPARATE

    def test_last_completed_step_none(self) -> None:
        state = PipelineState(video_id="abc123")
        assert state.last_completed_step is None

    def test_last_completed_step(self) -> None:
        state = PipelineState(video_id="abc123")
        for name in [StepName.DOWNLOAD, StepName.SEPARATE, StepName.TRANSCRIBE]:
            state.get_step(name).status = StepStatus.COMPLETED
        assert state.last_completed_step == StepName.TRANSCRIBE

    def test_all_steps_done(self) -> None:
        state = PipelineState(video_id="abc123")
        for name in STEP_ORDER:
            state.get_step(name).status = StepStatus.COMPLETED
        assert state.next_step is None

    def test_get_step_creates_on_demand(self) -> None:
        state = PipelineState(video_id="abc123")
        result = state.get_step(StepName.TRANSLATE)
        assert result.step == StepName.TRANSLATE
        assert result.status == StepStatus.PENDING

    def test_get_step_idempotent(self) -> None:
        """Calling get_step twice returns the same object."""
        state = PipelineState(video_id="abc123")
        r1 = state.get_step(StepName.DOWNLOAD)
        r1.status = StepStatus.RUNNING
        r2 = state.get_step(StepName.DOWNLOAD)
        assert r2.status == StepStatus.RUNNING
        assert r1 is r2

    def test_failed_step_blocks_next(self) -> None:
        """A failed step still appears as next_step (not completed)."""
        state = PipelineState(video_id="abc123")
        state.get_step(StepName.DOWNLOAD).status = StepStatus.COMPLETED
        state.get_step(StepName.SEPARATE).status = StepStatus.FAILED
        assert state.next_step == StepName.SEPARATE

    def test_default_target_language(self) -> None:
        state = PipelineState(video_id="abc")
        assert state.target_language == "ru"

    def test_serialization_roundtrip(self) -> None:
        state = PipelineState(
            video_id="abc123",
            url="https://youtube.com/watch?v=abc123",
            target_language="ru",
            meta=VideoMeta(video_id="abc123", title="Test", channel="Ch", duration=60.0),
            segments=[
                Segment(
                    id=0,
                    text="Hello",
                    start=0.0,
                    end=1.0,
                    words=[Word(text="Hello", start=0.0, end=0.5)],
                )
            ],
            speakers=[Speaker(id="SPEAKER_00", total_duration=10.0)],
        )
        state.get_step(StepName.DOWNLOAD).status = StepStatus.COMPLETED

        json_str = state.model_dump_json()
        restored = PipelineState.model_validate_json(json_str)

        assert restored.video_id == "abc123"
        assert restored.meta is not None
        assert restored.meta.title == "Test"
        assert len(restored.segments) == 1
        assert len(restored.segments[0].words) == 1
        assert restored.get_step(StepName.DOWNLOAD).status == StepStatus.COMPLETED

    def test_roundtrip_preserves_step_outputs(self) -> None:
        state = PipelineState(video_id="abc")
        step = state.get_step(StepName.DOWNLOAD)
        step.status = StepStatus.COMPLETED
        step.outputs = {"video": "video.mp4", "audio": "audio.wav"}

        restored = PipelineState.model_validate_json(state.model_dump_json())
        assert restored.get_step(StepName.DOWNLOAD).outputs["video"] == "video.mp4"



"""Tests for yt_dbl.pipeline.download — YouTube download step (mocked)."""

from __future__ import annotations

import json
import subprocess
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest

from yt_dbl.config import Settings
from yt_dbl.pipeline.base import DownloadError, StepValidationError
from yt_dbl.pipeline.download import DownloadStep
from yt_dbl.schemas import STEP_DIRS, PipelineState, StepName

if TYPE_CHECKING:
    from pathlib import Path

FAKE_META = {
    "title": "Test Video Title",
    "channel": "Test Channel",
    "uploader": "Test Uploader",
    "duration": 120.5,
    "id": "dQw4w9WgXcQ",
}


class TestDownloadStep:
    def _make_step(self, tmp_path: Path) -> tuple[DownloadStep, Settings, PipelineState]:
        cfg = Settings(work_dir=tmp_path / "work")
        step_dir = cfg.step_dir("test123", STEP_DIRS[StepName.DOWNLOAD])
        step = DownloadStep(settings=cfg, work_dir=step_dir)
        state = PipelineState(
            video_id="test123",
            url="https://youtube.com/watch?v=test123",
        )
        return step, cfg, state

    def test_validate_inputs_no_url(self, tmp_path: Path) -> None:
        step, _, state = self._make_step(tmp_path)
        state.url = ""
        with pytest.raises(StepValidationError, match="No URL"):
            step.validate_inputs(state)

    def test_validate_inputs_ok(self, tmp_path: Path) -> None:
        step, _, state = self._make_step(tmp_path)
        step.validate_inputs(state)  # Should not raise

    @patch("yt_dbl.pipeline.download.DownloadStep._download_video")
    @patch("yt_dbl.pipeline.download.DownloadStep._fetch_metadata")
    @patch("yt_dbl.pipeline.download.extract_audio")
    def test_run_success(
        self,
        mock_extract: MagicMock,
        mock_meta: MagicMock,
        mock_download: MagicMock,
        tmp_path: Path,
    ) -> None:
        step, _cfg, state = self._make_step(tmp_path)

        mock_meta.return_value = FAKE_META

        # Simulate file creation by download and extract
        def create_video(url: str, path: Path) -> None:
            path.write_bytes(b"fake-video-content")

        def create_audio(video_path: Path, audio_path: Path, **kwargs: object) -> Path:
            audio_path.write_bytes(b"fake-audio-content")
            return audio_path

        mock_download.side_effect = create_video
        mock_extract.side_effect = create_audio

        state = step.run(state)

        assert state.meta is not None
        assert state.meta.title == "Test Video Title"
        assert state.meta.channel == "Test Channel"
        assert state.meta.duration == 120.5

        result = state.get_step(StepName.DOWNLOAD)
        assert result.outputs["video"] == "video.mp4"
        assert result.outputs["audio"] == "audio.wav"

    @patch("yt_dbl.pipeline.download.DownloadStep._fetch_metadata")
    def test_fetch_metadata_timeout(
        self,
        mock_meta: MagicMock,
        tmp_path: Path,
    ) -> None:
        step, _, state = self._make_step(tmp_path)
        mock_meta.side_effect = DownloadError("Metadata fetch timed out (30s)")

        with pytest.raises(DownloadError, match="timed out"):
            step.run(state)

    @patch("yt_dbl.pipeline.download.DownloadStep._download_video")
    @patch("yt_dbl.pipeline.download.DownloadStep._fetch_metadata")
    def test_skips_existing_files(
        self,
        mock_meta: MagicMock,
        mock_download: MagicMock,
        tmp_path: Path,
    ) -> None:
        step, _cfg, state = self._make_step(tmp_path)
        mock_meta.return_value = FAKE_META

        # Pre-create files
        (step.step_dir / "video.mp4").write_bytes(b"existing-video")
        (step.step_dir / "audio.wav").write_bytes(b"existing-audio")

        state = step.run(state)

        mock_download.assert_not_called()
        assert state.meta is not None
        assert state.meta.title == "Test Video Title"


class TestFetchMetadata:
    def test_parses_json_output(self, tmp_path: Path) -> None:
        cfg = Settings(work_dir=tmp_path / "work")
        step_dir = cfg.step_dir("test123", STEP_DIRS[StepName.DOWNLOAD])
        step = DownloadStep(settings=cfg, work_dir=step_dir)

        mock_result = subprocess.CompletedProcess(
            args=[], returncode=0, stdout=json.dumps(FAKE_META), stderr=""
        )

        with patch("subprocess.run", return_value=mock_result):
            meta = step._fetch_metadata("https://youtube.com/watch?v=test")

        assert meta["title"] == "Test Video Title"
        assert meta["duration"] == 120.5

    def test_raises_on_failure(self, tmp_path: Path) -> None:
        cfg = Settings(work_dir=tmp_path / "work")
        step_dir = cfg.step_dir("test123", STEP_DIRS[StepName.DOWNLOAD])
        step = DownloadStep(settings=cfg, work_dir=step_dir)

        with (
            patch(
                "subprocess.run",
                side_effect=subprocess.CalledProcessError(1, "yt-dlp", stderr="Video unavailable"),
            ),
            pytest.raises(DownloadError, match="Failed to fetch metadata"),
        ):
            step._fetch_metadata("https://youtube.com/watch?v=invalid")


class TestDownloadVideo:
    """Tests for _download_video error handling."""

    @patch("subprocess.Popen")
    def test_yt_dlp_not_found(self, mock_popen: MagicMock, tmp_path: Path) -> None:
        mock_popen.side_effect = FileNotFoundError()
        cfg = Settings(work_dir=tmp_path / "work")
        step_dir = cfg.step_dir("test123", STEP_DIRS[StepName.DOWNLOAD])
        step = DownloadStep(settings=cfg, work_dir=step_dir)

        with pytest.raises(DownloadError, match="yt-dlp not found"):
            step._download_video("https://youtube.com/watch?v=x", step_dir / "video.mp4")

    @patch("subprocess.Popen")
    def test_nonzero_exit_code(self, mock_popen: MagicMock, tmp_path: Path) -> None:
        proc = MagicMock()
        proc.stdout = iter(["[download] 100% done\n"])
        proc.wait.return_value = 1
        mock_popen.return_value = proc

        cfg = Settings(work_dir=tmp_path / "work")
        step_dir = cfg.step_dir("test123", STEP_DIRS[StepName.DOWNLOAD])
        step = DownloadStep(settings=cfg, work_dir=step_dir)

        with pytest.raises(DownloadError, match="exited with code 1"):
            step._download_video("https://youtube.com/watch?v=x", step_dir / "video.mp4")

    @patch("subprocess.Popen")
    def test_kills_process_on_unexpected_error(self, mock_popen: MagicMock, tmp_path: Path) -> None:
        """If an unexpected error occurs while reading stdout, yt-dlp is killed."""
        proc = MagicMock()
        proc.stdout = iter(["boom"])
        # Simulate the progress bar raising an unexpected error
        proc.wait.side_effect = RuntimeError("unexpected")
        mock_popen.return_value = proc

        cfg = Settings(work_dir=tmp_path / "work")
        step_dir = cfg.step_dir("test123", STEP_DIRS[StepName.DOWNLOAD])
        step = DownloadStep(settings=cfg, work_dir=step_dir)

        with pytest.raises(RuntimeError, match="unexpected"):
            step._download_video("https://youtube.com/watch?v=x", step_dir / "video.mp4")

        proc.kill.assert_called_once()


class TestDownloadStepOutputVerification:
    """Verify that run() checks output presence."""

    @patch("yt_dbl.pipeline.download.DownloadStep._download_video")
    @patch("yt_dbl.pipeline.download.DownloadStep._fetch_metadata")
    @patch("yt_dbl.pipeline.download.extract_audio")
    def test_missing_video_raises(
        self,
        mock_extract: MagicMock,
        mock_meta: MagicMock,
        mock_download: MagicMock,
        tmp_path: Path,
    ) -> None:
        cfg = Settings(work_dir=tmp_path / "work")
        step_dir = cfg.step_dir("test123", STEP_DIRS[StepName.DOWNLOAD])
        step = DownloadStep(settings=cfg, work_dir=step_dir)
        state = PipelineState(video_id="test123", url="https://youtube.com/watch?v=test123")

        mock_meta.return_value = FAKE_META
        # download does nothing — video file won't exist
        mock_download.side_effect = lambda *a, **kw: None
        mock_extract.side_effect = lambda *a, **kw: None

        with pytest.raises(DownloadError, match="Video file was not created"):
            step.run(state)



"""Tests for pipeline runner — checkpointing and resume."""

import json
from pathlib import Path
from typing import Any
from unittest.mock import MagicMock, patch

import pytest

from tests.conftest import prefill_download
from yt_dbl.config import Settings
from yt_dbl.pipeline.base import StepValidationError
from yt_dbl.pipeline.runner import PipelineRunner, load_state, save_state
from yt_dbl.schemas import STEP_DIRS, PipelineState, Segment, StepName, StepStatus


def _fake_separation_factory(sep_dir: Path) -> Any:
    """Return a side_effect callable that creates fake separation outputs."""

    def _fake(audio_path: Path) -> None:
        (sep_dir / "vocals.wav").write_bytes(b"fake-vocals")
        (sep_dir / "background.wav").write_bytes(b"fake-background")

    return _fake


def _fake_transcription_factory() -> dict[str, Any]:
    """Return mock side_effects for _run_asr and _run_alignment."""
    from yt_dbl.schemas import Segment, Word

    fake_asr_segments = [
        {"start": 0.0, "end": 3.0, "speaker_id": 0, "text": "Hello world."},
        {"start": 4.0, "end": 6.0, "speaker_id": 1, "text": "Goodbye."},
    ]

    fake_segments = [
        Segment(
            id=0,
            text="Hello world.",
            start=0.0,
            end=3.0,
            speaker="SPEAKER_00",
            language="en",
            words=[Word(text="Hello", start=0.0, end=1.5), Word(text="world.", start=1.5, end=3.0)],
        ),
        Segment(
            id=1,
            text="Goodbye.",
            start=4.0,
            end=6.0,
            speaker="SPEAKER_01",
            language="en",
            words=[Word(text="Goodbye.", start=4.0, end=6.0)],
        ),
    ]

    def _fake_asr(vocals_path: Path) -> list[dict[str, Any]]:
        return fake_asr_segments

    def _fake_alignment(
        vocals_path: Path,
        raw_segments: list[dict[str, Any]],
        detected_lang: str = "en",
    ) -> list[Segment]:
        return fake_segments

    return {
        "asr": _fake_asr,
        "alignment": _fake_alignment,
    }


def _fake_translation(
    segments: list[Segment],
    target_language: str,
    source_language: str = "auto-detected",
) -> dict[int, str]:
    """Fake translate that returns deterministic translations."""
    return {seg.id: f"[{target_language}] {seg.text}" for seg in segments}


def _fake_postprocess_segment(input_path: Path, output_path: Path, **_kwargs: Any) -> Path:
    """Mock postprocess_segment that creates the output file."""
    output_path.parent.mkdir(parents=True, exist_ok=True)
    output_path.write_bytes(b"fake-audio")
    return output_path


def _fake_tts_model() -> Any:
    """Return a fake TTS model whose generate() returns dummy audio."""
    import numpy as np

    model = MagicMock()
    result = MagicMock()
    result.audio = np.zeros(12000, dtype=np.float32)
    result.sample_rate = 12000
    model.generate.return_value = [result]
    return model


def _pipeline_patches(sep_dir: Path) -> Any:
    """Context manager stack for all pipeline mocks (separate+transcribe+translate+synthesize)."""
    from contextlib import ExitStack

    fakes = _fake_transcription_factory()

    stack = ExitStack()
    stack.enter_context(
        patch(
            "yt_dbl.pipeline.separate.SeparateStep._run_separation",
            side_effect=_fake_separation_factory(sep_dir),
        )
    )
    stack.enter_context(
        patch(
            "yt_dbl.pipeline.transcribe.TranscribeStep._run_asr",
            side_effect=fakes["asr"],
        )
    )
    stack.enter_context(
        patch(
            "yt_dbl.pipeline.transcribe.TranscribeStep._run_alignment",
            side_effect=fakes["alignment"],
        )
    )
    stack.enter_context(
        patch(
            "yt_dbl.pipeline.translate.TranslateStep._translate_all",
            side_effect=_fake_translation,
        )
    )
    # Synthesize mocks: skip real TTS + ffmpeg
    stack.enter_context(
        patch(
            "yt_dbl.pipeline.synthesize.SynthesizeStep._load_tts_model",
            return_value=_fake_tts_model(),
        )
    )
    stack.enter_context(
        patch(
            "yt_dbl.pipeline.synthesize.SynthesizeStep._run_tts",
            return_value=__import__("numpy").zeros(12000, dtype=__import__("numpy").float32),
        )
    )
    stack.enter_context(
        patch("yt_dbl.utils.audio_processing.run_ffmpeg"),
    )
    stack.enter_context(
        patch(
            "yt_dbl.pipeline.synthesize.postprocess_segment",
            side_effect=_fake_postprocess_segment,
        ),
    )
    stack.enter_context(
        patch(
            "yt_dbl.pipeline.synthesize.get_audio_duration",
            return_value=1.0,
        ),
    )
    stack.enter_context(
        patch("yt_dbl.pipeline.synthesize.SynthesizeStep._save_wav"),
    )
    # Assemble mocks: skip real ffmpeg for final muxing
    stack.enter_context(
        patch("yt_dbl.pipeline.assemble.run_ffmpeg"),
    )
    return stack


class TestCheckpoints:
    def test_save_and_load(self, tmp_path: Path) -> None:
        cfg = Settings(work_dir=tmp_path / "work")
        state = PipelineState(video_id="test123", url="https://example.com", target_language="ru")
        state.get_step(StepName.DOWNLOAD).status = StepStatus.COMPLETED

        save_state(state, cfg)
        loaded = load_state(cfg, "test123")

        assert loaded is not None
        assert loaded.video_id == "test123"
        assert loaded.get_step(StepName.DOWNLOAD).status == StepStatus.COMPLETED

    def test_load_nonexistent(self, tmp_path: Path) -> None:
        cfg = Settings(work_dir=tmp_path / "work")
        assert load_state(cfg, "nonexistent") is None

    def test_state_file_is_valid_json(self, tmp_path: Path) -> None:
        cfg = Settings(work_dir=tmp_path / "work")
        state = PipelineState(video_id="test123")
        path = save_state(state, cfg)

        data = json.loads(path.read_text())
        assert data["video_id"] == "test123"

    def test_save_no_leftover_tmp_files(self, tmp_path: Path) -> None:
        """After save_state completes, no .tmp files remain."""
        cfg = Settings(work_dir=tmp_path / "work")
        state = PipelineState(video_id="test123")
        save_state(state, cfg)

        job_dir = cfg.work_dir / "test123"
        tmp_files = list(job_dir.glob("*.tmp"))
        assert tmp_files == []

    def test_save_overwrites_atomically(self, tmp_path: Path) -> None:
        """A second save_state replaces the first without corruption."""
        cfg = Settings(work_dir=tmp_path / "work")
        state = PipelineState(video_id="test123", target_language="en")
        save_state(state, cfg)

        state.target_language = "fr"
        save_state(state, cfg)

        loaded = load_state(cfg, "test123")
        assert loaded is not None
        assert loaded.target_language == "fr"

    def test_save_cleans_tmp_on_error(self, tmp_path: Path) -> None:
        """If serialisation fails, no .tmp file is left behind."""
        cfg = Settings(work_dir=tmp_path / "work")
        state = PipelineState(video_id="test123")
        # Ensure job dir exists
        cfg.job_dir("test123")

        with (
            patch.object(
                PipelineState,
                "model_dump_json",
                side_effect=RuntimeError("boom"),
            ),
            pytest.raises(RuntimeError, match="boom"),
        ):
            save_state(state, cfg)

        job_dir = cfg.work_dir / "test123"
        tmp_files = list(job_dir.glob("*.tmp"))
        assert tmp_files == []


class TestPipelineRunner:
    def test_pipeline_from_separate(self, tmp_path: Path) -> None:
        """Run pipeline skipping download (which needs real yt-dlp)."""
        cfg = Settings(work_dir=tmp_path / "work", anthropic_api_key="sk-test")
        state = PipelineState(
            video_id="test123",
            url="https://youtube.com/watch?v=test123",
            target_language="ru",
        )
        state = prefill_download(state, cfg)
        save_state(state, cfg)

        sep_dir = cfg.step_dir("test123", STEP_DIRS[StepName.SEPARATE])

        runner = PipelineRunner(cfg)
        with _pipeline_patches(sep_dir):
            state = runner.run(state)

        assert state.next_step is None
        for step_name in [
            StepName.SEPARATE,
            StepName.TRANSCRIBE,
            StepName.TRANSLATE,
            StepName.SYNTHESIZE,
            StepName.ASSEMBLE,
        ]:
            assert state.get_step(step_name).status == StepStatus.COMPLETED

    def test_resume_skips_completed(self, tmp_path: Path) -> None:
        cfg = Settings(work_dir=tmp_path / "work", anthropic_api_key="sk-test")
        state = PipelineState(video_id="test123", url="https://example.com")
        state = prefill_download(state, cfg)

        sep_dir = cfg.step_dir("test123", STEP_DIRS[StepName.SEPARATE])

        runner = PipelineRunner(cfg)
        with _pipeline_patches(sep_dir):
            state = runner.run(state)

        # Mark last 3 as pending again
        for name in [StepName.TRANSLATE, StepName.SYNTHESIZE, StepName.ASSEMBLE]:
            state.get_step(name).status = StepStatus.PENDING
        save_state(state, cfg)

        loaded = load_state(cfg, "test123")
        assert loaded is not None
        assert loaded.next_step == StepName.TRANSLATE

    def test_from_step(self, tmp_path: Path) -> None:
        cfg = Settings(work_dir=tmp_path / "work", anthropic_api_key="sk-test")
        state = PipelineState(video_id="test123", url="https://example.com")
        state = prefill_download(state, cfg)

        sep_dir = cfg.step_dir("test123", STEP_DIRS[StepName.SEPARATE])

        runner = PipelineRunner(cfg)
        with _pipeline_patches(sep_dir):
            state = runner.run(state)
            state = runner.run(state, from_step=StepName.TRANSLATE)

        assert state.get_step(StepName.TRANSLATE).status == StepStatus.COMPLETED

    def test_step_failure_stops_pipeline(self, tmp_path: Path) -> None:
        """When a step raises, its status becomes FAILED and pipeline stops."""
        cfg = Settings(work_dir=tmp_path / "work", anthropic_api_key="sk-test")
        state = PipelineState(video_id="test123", url="https://example.com")
        state = prefill_download(state, cfg)

        runner = PipelineRunner(cfg)

        # Patch SeparateStep.run to raise
        with patch(
            "yt_dbl.pipeline.separate.SeparateStep.run",
            side_effect=RuntimeError("boom"),
        ):
            state = runner.run(state)

        assert state.get_step(StepName.SEPARATE).status == StepStatus.FAILED
        assert "boom" in state.get_step(StepName.SEPARATE).error
        # Steps after failure should still be pending
        assert state.get_step(StepName.TRANSCRIBE).status == StepStatus.PENDING

    def test_checkpoint_saved_on_success(self, tmp_path: Path) -> None:
        """After each step completes, state is persisted to disk."""
        cfg = Settings(work_dir=tmp_path / "work", anthropic_api_key="sk-test")
        state = PipelineState(video_id="test123", url="https://example.com")
        state = prefill_download(state, cfg)

        sep_dir = cfg.step_dir("test123", STEP_DIRS[StepName.SEPARATE])

        runner = PipelineRunner(cfg)
        with _pipeline_patches(sep_dir):
            state = runner.run(state)

        loaded = load_state(cfg, "test123")
        assert loaded is not None
        assert loaded.get_step(StepName.ASSEMBLE).status == StepStatus.COMPLETED

    def test_checkpoint_saved_on_failure(self, tmp_path: Path) -> None:
        """Failed state should also be persisted."""
        cfg = Settings(work_dir=tmp_path / "work", anthropic_api_key="sk-test")
        state = PipelineState(video_id="test123", url="https://example.com")
        state = prefill_download(state, cfg)

        runner = PipelineRunner(cfg)

        with patch(
            "yt_dbl.pipeline.separate.SeparateStep.run",
            side_effect=RuntimeError("disk full"),
        ):
            state = runner.run(state)

        loaded = load_state(cfg, "test123")
        assert loaded is not None
        assert loaded.get_step(StepName.SEPARATE).status == StepStatus.FAILED
        assert "disk full" in loaded.get_step(StepName.SEPARATE).error

    def test_step_timing_recorded(self, tmp_path: Path) -> None:
        """Completed steps record timing metadata."""
        cfg = Settings(work_dir=tmp_path / "work", anthropic_api_key="sk-test")
        state = PipelineState(video_id="test123", url="https://example.com")
        state = prefill_download(state, cfg)

        sep_dir = cfg.step_dir("test123", STEP_DIRS[StepName.SEPARATE])

        runner = PipelineRunner(cfg)
        with _pipeline_patches(sep_dir):
            state = runner.run(state)

        result = state.get_step(StepName.SEPARATE)
        assert result.started_at != ""
        assert result.finished_at != ""
        assert result.duration_sec >= 0.0


class TestEarlyValidation:
    def test_missing_api_key_raises(self, tmp_path: Path) -> None:
        """Pipeline refuses to start when API key is missing and translate hasn't run."""
        cfg = Settings(work_dir=tmp_path / "work", anthropic_api_key="")
        state = PipelineState(video_id="test123", url="https://example.com")
        state = prefill_download(state, cfg)

        runner = PipelineRunner(cfg)
        with pytest.raises(StepValidationError, match="Anthropic API key"):
            runner.run(state)

    def test_api_key_not_needed_when_translate_done(self, tmp_path: Path) -> None:
        """If translate is already COMPLETED, no API key is required."""
        cfg = Settings(work_dir=tmp_path / "work", anthropic_api_key="")
        state = PipelineState(video_id="test123", url="https://example.com")
        state = prefill_download(state, cfg)
        state.get_step(StepName.TRANSLATE).status = StepStatus.COMPLETED

        sep_dir = cfg.step_dir("test123", STEP_DIRS[StepName.SEPARATE])
        runner = PipelineRunner(cfg)
        with _pipeline_patches(sep_dir):
            state = runner.run(state)
        # Pipeline should run without raising


class TestModelCleanup:
    def test_models_unloaded_on_keyboard_interrupt(self, tmp_path: Path) -> None:
        """Models are freed even if a step raises KeyboardInterrupt."""
        cfg = Settings(work_dir=tmp_path / "work", anthropic_api_key="sk-test")
        state = PipelineState(video_id="test123", url="https://example.com")
        state = prefill_download(state, cfg)

        runner = PipelineRunner(cfg)

        with (
            patch.object(runner.model_manager, "unload_all") as mock_unload,
            patch(
                "yt_dbl.pipeline.separate.SeparateStep.run",
                side_effect=KeyboardInterrupt,
            ),
            pytest.raises(KeyboardInterrupt),
        ):
            runner.run(state)

        mock_unload.assert_called_once()



"""Tests for CLI commands."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest
from typer.testing import CliRunner

from yt_dbl.cli import _extract_video_id, _invalidate_language_steps, _step_name_from_str, app
from yt_dbl.schemas import STEP_DIRS, PipelineState, StepName, StepStatus

if TYPE_CHECKING:
    from pathlib import Path

runner = CliRunner()

_ANSI_RE = re.compile(r"\x1b\[[0-9;]*m")


def _plain(text: str) -> str:
    """Strip ANSI escape codes so assertions survive Rich markup."""
    return _ANSI_RE.sub("", text)


class TestCLI:
    def test_version(self) -> None:
        result = runner.invoke(app, ["--version"])
        assert result.exit_code == 0
        assert re.search(r"\d+\.\d+\.\d+", result.output.strip())

    def test_help(self) -> None:
        result = runner.invoke(app, ["--help"])
        assert result.exit_code == 0
        assert "dub" in result.output
        assert "resume" in result.output
        assert "models" in result.output

    def test_dub_help(self) -> None:
        result = runner.invoke(app, ["dub", "--help"])
        assert result.exit_code == 0
        output = _plain(result.output)
        assert "--target-language" in output
        assert "--from-step" in output

    def test_resume_nonexistent(self) -> None:
        result = runner.invoke(app, ["resume", "nonexistent_video_id_xyz"])
        assert result.exit_code == 1

    def test_status_nonexistent(self) -> None:
        result = runner.invoke(app, ["status", "nonexistent_video_id_xyz"])
        assert result.exit_code == 1

    def test_models_list(self) -> None:
        result = runner.invoke(app, ["models", "list"])
        assert result.exit_code == 0

    def test_models_download(self) -> None:
        with patch(
            "yt_dbl.models.registry.check_model_downloaded",
            return_value=True,
        ):
            result = runner.invoke(app, ["models", "download"])
        assert result.exit_code == 0


class TestExtractVideoId:
    """Unit tests for _extract_video_id helper."""

    @pytest.mark.parametrize(
        ("url", "expected"),
        [
            ("https://www.youtube.com/watch?v=dQw4w9WgXcQ", "dQw4w9WgXcQ"),
            ("https://youtu.be/dQw4w9WgXcQ", "dQw4w9WgXcQ"),
            ("https://youtube.com/shorts/dQw4w9WgXcQ", "dQw4w9WgXcQ"),
            ("https://youtube.com/v/dQw4w9WgXcQ", "dQw4w9WgXcQ"),
            ("https://www.youtube.com/watch?v=abcdefghijk&list=PL123", "abcdefghijk"),
        ],
        ids=["standard", "short", "shorts", "embed", "extra-params"],
    )
    def test_extracts_id(self, url: str, expected: str) -> None:
        assert _extract_video_id(url) == expected

    def test_fallback_sanitizes(self) -> None:
        """Non-YouTube URL should be sanitized as fallback."""
        vid = _extract_video_id("https://example.com/some-video")
        assert len(vid) <= 64
        assert all(c.isalnum() or c in "-_" for c in vid)


class TestStepNameFromStr:
    @pytest.mark.parametrize(
        ("value", "expected"),
        [
            ("download", StepName.DOWNLOAD),
            ("TRANSLATE", StepName.TRANSLATE),
            ("separate", StepName.SEPARATE),
        ],
        ids=["lowercase", "uppercase", "another-step"],
    )
    def test_valid_step(self, value: str, expected: StepName) -> None:
        assert _step_name_from_str(value) == expected

    def test_invalid_step(self) -> None:
        import typer

        with pytest.raises(typer.BadParameter, match="nonexistent"):
            _step_name_from_str("nonexistent")


class TestDubCommand:
    """Integration tests for the `dub` command (mocked pipeline)."""

    @patch("yt_dbl.pipeline.runner.PipelineRunner")
    @patch("yt_dbl.pipeline.runner.save_state")
    @patch("yt_dbl.pipeline.runner.load_state", return_value=None)
    def test_dub_new_job(
        self,
        mock_load: MagicMock,
        mock_save: MagicMock,
        mock_runner_cls: MagicMock,
        tmp_path: Path,
    ) -> None:
        mock_runner = MagicMock()
        mock_runner_cls.return_value = mock_runner

        result = runner.invoke(app, ["dub", "https://youtube.com/watch?v=dQw4w9WgXcQ"])

        assert result.exit_code == 0
        mock_runner.run.assert_called_once()

    @patch("yt_dbl.pipeline.runner.PipelineRunner")
    @patch("yt_dbl.pipeline.runner.save_state")
    @patch("yt_dbl.pipeline.runner.load_state")
    def test_dub_resumes_existing(
        self,
        mock_load: MagicMock,
        mock_save: MagicMock,
        mock_runner_cls: MagicMock,
    ) -> None:
        existing_state = PipelineState(
            video_id="dQw4w9WgXcQ",
            url="https://youtube.com/watch?v=dQw4w9WgXcQ",
        )
        existing_state.get_step(StepName.DOWNLOAD).status = StepStatus.COMPLETED
        mock_load.return_value = existing_state

        mock_runner = MagicMock()
        mock_runner_cls.return_value = mock_runner

        result = runner.invoke(app, ["dub", "https://youtube.com/watch?v=dQw4w9WgXcQ"])

        assert result.exit_code == 0
        # Should NOT create a new state — load returned existing
        mock_save.assert_not_called()

    @patch("yt_dbl.pipeline.runner.PipelineRunner")
    @patch("yt_dbl.pipeline.runner.save_state")
    @patch("yt_dbl.pipeline.runner.load_state", return_value=None)
    def test_dub_with_from_step(
        self,
        mock_load: MagicMock,
        mock_save: MagicMock,
        mock_runner_cls: MagicMock,
    ) -> None:
        mock_runner = MagicMock()
        mock_runner_cls.return_value = mock_runner

        result = runner.invoke(
            app,
            ["dub", "https://youtube.com/watch?v=abc12345678", "--from-step", "translate"],
        )

        assert result.exit_code == 0
        call_kwargs = mock_runner.run.call_args
        assert call_kwargs[1]["from_step"] == StepName.TRANSLATE


class TestResumeCommand:
    @patch("yt_dbl.pipeline.runner.PipelineRunner")
    @patch("yt_dbl.pipeline.runner.load_state")
    def test_resume_existing_job(
        self,
        mock_load: MagicMock,
        mock_runner_cls: MagicMock,
    ) -> None:
        state = PipelineState(video_id="test123", url="https://example.com")
        state.get_step(StepName.DOWNLOAD).status = StepStatus.COMPLETED
        mock_load.return_value = state

        mock_runner = MagicMock()
        mock_runner_cls.return_value = mock_runner

        result = runner.invoke(app, ["resume", "test123"])

        assert result.exit_code == 0
        mock_runner.run.assert_called_once()

    @patch("yt_dbl.pipeline.runner.load_state")
    def test_resume_completed_job(self, mock_load: MagicMock) -> None:
        state = PipelineState(video_id="test123", url="https://example.com")
        for step in StepName:
            state.get_step(step).status = StepStatus.COMPLETED
        mock_load.return_value = state

        result = runner.invoke(app, ["resume", "test123"])
        assert result.exit_code == 0
        assert "completed" in result.output.lower()


class TestStatusCommand:
    def test_status_shows_table(self, tmp_path: Path) -> None:
        from yt_dbl.schemas import VideoMeta

        state = PipelineState(
            video_id="test123",
            url="https://example.com",
            meta=VideoMeta(video_id="test123", title="Test Title", channel="Ch", duration=60.0),
        )
        state.get_step(StepName.DOWNLOAD).status = StepStatus.COMPLETED
        state.get_step(StepName.DOWNLOAD).duration_sec = 5.3
        state.get_step(StepName.SEPARATE).status = StepStatus.FAILED
        state.get_step(StepName.SEPARATE).error = "Some error happened during separation"

        # Write a state file
        job_dir = tmp_path / "test123"
        job_dir.mkdir(parents=True)
        state_file = job_dir / "state.json"
        state_file.write_text(state.model_dump_json(indent=2))

        # Make load_state find it
        with patch("yt_dbl.pipeline.runner.load_state", return_value=state):
            result = runner.invoke(app, ["status", "test123"])

        assert result.exit_code == 0
        assert "completed" in result.output.lower()
        assert "failed" in result.output.lower()
        assert "Test Title" in result.output


class TestInvalidateLanguageSteps:
    """Unit tests for _invalidate_language_steps helper."""

    def test_resets_completed_steps(self, tmp_path: Path) -> None:
        """Completed translate/synthesize/assemble steps are reset to PENDING."""
        from yt_dbl.config import Settings

        cfg = Settings(work_dir=tmp_path)
        state = PipelineState(video_id="vid1", target_language="en")
        for step in (StepName.TRANSLATE, StepName.SYNTHESIZE, StepName.ASSEMBLE):
            r = state.get_step(step)
            r.status = StepStatus.COMPLETED
            r.outputs = {"key": "value"}
            r.duration_sec = 10.0
            r.error = "old error"

        _invalidate_language_steps(state, cfg, "vid1")

        for step in (StepName.TRANSLATE, StepName.SYNTHESIZE, StepName.ASSEMBLE):
            r = state.get_step(step)
            assert r.status == StepStatus.PENDING
            assert r.outputs == {}
            assert r.duration_sec == 0.0
            assert r.error == ""

    def test_removes_cached_step_dirs(self, tmp_path: Path) -> None:
        """Cached step directories are deleted."""
        from yt_dbl.config import Settings

        cfg = Settings(work_dir=tmp_path)
        state = PipelineState(video_id="vid1")

        # Create step directories with dummy files
        for step in (StepName.TRANSLATE, StepName.SYNTHESIZE, StepName.ASSEMBLE):
            d = tmp_path / "vid1" / STEP_DIRS[step]
            d.mkdir(parents=True)
            (d / "dummy.json").write_text("{}")

        _invalidate_language_steps(state, cfg, "vid1")

        for step in (StepName.TRANSLATE, StepName.SYNTHESIZE, StepName.ASSEMBLE):
            assert not (tmp_path / "vid1" / STEP_DIRS[step]).exists()

    def test_removes_result_files(self, tmp_path: Path) -> None:
        """Result mp4/mkv files are deleted from job root."""
        from yt_dbl.config import Settings

        cfg = Settings(work_dir=tmp_path)
        state = PipelineState(video_id="vid1")
        job_dir = tmp_path / "vid1"
        job_dir.mkdir(parents=True)
        (job_dir / "result.mp4").write_text("fake")
        (job_dir / "result.mkv").write_text("fake")

        _invalidate_language_steps(state, cfg, "vid1")

        assert not (job_dir / "result.mp4").exists()
        assert not (job_dir / "result.mkv").exists()

    def test_leaves_earlier_steps_untouched(self, tmp_path: Path) -> None:
        """Download / separate / transcribe steps are NOT affected."""
        from yt_dbl.config import Settings

        cfg = Settings(work_dir=tmp_path)
        state = PipelineState(video_id="vid1")
        for step in (StepName.DOWNLOAD, StepName.SEPARATE, StepName.TRANSCRIBE):
            r = state.get_step(step)
            r.status = StepStatus.COMPLETED
            r.outputs = {"key": "value"}

        _invalidate_language_steps(state, cfg, "vid1")

        for step in (StepName.DOWNLOAD, StepName.SEPARATE, StepName.TRANSCRIBE):
            r = state.get_step(step)
            assert r.status == StepStatus.COMPLETED
            assert r.outputs == {"key": "value"}


class TestDubLanguageOverride:
    """Integration tests for --target-language override on existing jobs."""

    @patch("yt_dbl.pipeline.runner.PipelineRunner")
    @patch("yt_dbl.pipeline.runner.save_state")
    @patch("yt_dbl.pipeline.runner.load_state")
    def test_language_override_updates_state(
        self,
        mock_load: MagicMock,
        mock_save: MagicMock,
        mock_runner_cls: MagicMock,
    ) -> None:
        """Passing --target-language on existing job updates state."""
        existing = PipelineState(
            video_id="dQw4w9WgXcQ",
            url="https://youtube.com/watch?v=dQw4w9WgXcQ",
            target_language="ru",
        )
        existing.get_step(StepName.TRANSLATE).status = StepStatus.COMPLETED
        mock_load.return_value = existing

        mock_runner = MagicMock()
        mock_runner_cls.return_value = mock_runner

        result = runner.invoke(
            app,
            ["dub", "https://youtube.com/watch?v=dQw4w9WgXcQ", "-t", "fr"],
        )

        assert result.exit_code == 0
        # State should have been saved with new language
        mock_save.assert_called_once()
        saved_state = mock_save.call_args[0][0]
        assert saved_state.target_language == "fr"
        # Translate step should be invalidated
        assert saved_state.get_step(StepName.TRANSLATE).status == StepStatus.PENDING

    @patch("yt_dbl.pipeline.runner.PipelineRunner")
    @patch("yt_dbl.pipeline.runner.save_state")
    @patch("yt_dbl.pipeline.runner.load_state")
    def test_same_language_no_override(
        self,
        mock_load: MagicMock,
        mock_save: MagicMock,
        mock_runner_cls: MagicMock,
    ) -> None:
        """Passing same --target-language does not trigger invalidation."""
        existing = PipelineState(
            video_id="dQw4w9WgXcQ",
            url="https://youtube.com/watch?v=dQw4w9WgXcQ",
            target_language="ru",
        )
        existing.get_step(StepName.TRANSLATE).status = StepStatus.COMPLETED
        mock_load.return_value = existing

        mock_runner = MagicMock()
        mock_runner_cls.return_value = mock_runner

        result = runner.invoke(
            app,
            ["dub", "https://youtube.com/watch?v=dQw4w9WgXcQ", "-t", "ru"],
        )

        assert result.exit_code == 0
        mock_save.assert_not_called()

    @patch("yt_dbl.pipeline.runner.PipelineRunner")
    @patch("yt_dbl.pipeline.runner.save_state")
    @patch("yt_dbl.pipeline.runner.load_state")
    def test_no_language_flag_no_override(
        self,
        mock_load: MagicMock,
        mock_save: MagicMock,
        mock_runner_cls: MagicMock,
    ) -> None:
        """Omitting --target-language preserves saved state."""
        existing = PipelineState(
            video_id="dQw4w9WgXcQ",
            url="https://youtube.com/watch?v=dQw4w9WgXcQ",
            target_language="fr",
        )
        mock_load.return_value = existing

        mock_runner = MagicMock()
        mock_runner_cls.return_value = mock_runner

        result = runner.invoke(
            app,
            ["dub", "https://youtube.com/watch?v=dQw4w9WgXcQ"],
        )

        assert result.exit_code == 0
        mock_save.assert_not_called()
        assert existing.target_language == "fr"



"""Tests for yt_dbl.config — settings, paths, and RAM-based auto-detection."""

from __future__ import annotations

import subprocess
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import pytest

if TYPE_CHECKING:
    from pathlib import Path

from yt_dbl.config import (
    Settings,
    _detect_max_models,
    _detect_separation_batch_size,
    _get_total_ram_gb,
    _macos_ram_gb,
)


class TestSettings:
    def test_defaults(self) -> None:
        s = Settings(anthropic_api_key="test", hf_token="test")
        assert s.target_language == "ru"
        assert s.output_format == "mp4"
        assert s.subtitle_mode == "softsub"
        assert s.background_volume == 0.15
        assert s.max_speed_factor == 1.4
        assert s.voice_ref_duration == 7.0
        assert s.max_loaded_models >= 1  # auto-detected based on RAM
        assert s.separation_batch_size >= 1  # auto-detected based on RAM
        assert s.sample_rate == 48000
        assert s.claude_model == "claude-sonnet-4-5"

    def test_max_loaded_models_auto(self) -> None:
        """Default 0 triggers auto-detection → always ≥ 1."""
        s = Settings()
        assert s.max_loaded_models >= 1

    def test_max_loaded_models_explicit(self) -> None:
        """Explicit value overrides auto-detection."""
        s = Settings(max_loaded_models=5)
        assert s.max_loaded_models == 5

    def test_separation_batch_size_auto(self) -> None:
        """Default 0 triggers auto-detection → always ≥ 1."""
        s = Settings()
        assert s.separation_batch_size >= 1

    def test_separation_batch_size_explicit(self) -> None:
        """Explicit value overrides auto-detection."""
        s = Settings(separation_batch_size=3)
        assert s.separation_batch_size == 3

    def test_job_dir(self, tmp_path: Path) -> None:
        s = Settings(work_dir=tmp_path / "work")
        d = s.job_dir("abc123")
        assert d.exists()
        assert d.name == "abc123"

    def test_step_dir(self, tmp_path: Path) -> None:
        s = Settings(work_dir=tmp_path / "work")
        d = s.step_dir("abc123", "01_download")
        assert d.exists()
        assert d.name == "01_download"
        assert d.parent.name == "abc123"

    def test_background_volume_validation(self) -> None:
        with pytest.raises(ValueError, match="less than or equal to 1"):
            Settings(background_volume=1.5)

    def test_max_speed_validation(self) -> None:
        with pytest.raises(ValueError, match="greater than or equal to 1"):
            Settings(max_speed_factor=0.5)

    def test_voice_ref_duration_validation(self) -> None:
        with pytest.raises(ValueError, match="greater than or equal to 3"):
            Settings(voice_ref_duration=1.0)
        with pytest.raises(ValueError, match="less than or equal to 30"):
            Settings(voice_ref_duration=60.0)


# ── RAM-based auto-detection ────────────────────────────────────────────────


class TestDetectMaxModels:
    @pytest.mark.parametrize(
        ("ram_gb", "expected"),
        [
            (0.0, 1),
            (8.0, 1),
            (16.0, 1),
            (17.0, 2),
            (24.0, 2),
            (31.0, 2),
            (32.0, 3),
            (64.0, 3),
        ],
        ids=["zero", "8gb", "16gb", "17gb", "24gb", "31gb", "32gb", "64gb"],
    )
    @patch("yt_dbl.config._get_total_ram_gb")
    def test_ram_tiers(self, mock_ram: MagicMock, ram_gb: float, expected: int) -> None:
        mock_ram.return_value = ram_gb
        assert _detect_max_models() == expected


class TestDetectSeparationBatchSize:
    @pytest.mark.parametrize(
        ("ram_gb", "expected"),
        [
            (8.0, 1),
            (16.0, 1),
            (17.0, 2),
            (31.0, 2),
            (32.0, 4),
            (47.0, 4),
            (48.0, 8),
            (128.0, 8),
        ],
        ids=["8gb", "16gb", "17gb", "31gb", "32gb", "47gb", "48gb", "128gb"],
    )
    @patch("yt_dbl.config._get_total_ram_gb")
    def test_ram_tiers(self, mock_ram: MagicMock, ram_gb: float, expected: int) -> None:
        mock_ram.return_value = ram_gb
        assert _detect_separation_batch_size() == expected


class TestGetTotalRAM:
    @patch("os.sysconf")
    def test_via_sysconf(self, mock_sysconf: MagicMock) -> None:
        # 4M pages * 4096 bytes = 16 GiB
        mock_sysconf.side_effect = lambda name: {"SC_PAGE_SIZE": 4096, "SC_PHYS_PAGES": 4194304}[
            name
        ]
        assert _get_total_ram_gb() == pytest.approx(16.0)

    @patch("yt_dbl.config._macos_ram_gb", return_value=32.0)
    @patch("os.sysconf", side_effect=AttributeError)
    def test_fallback_to_macos(self, mock_sysconf: MagicMock, mock_macos: MagicMock) -> None:
        result = _get_total_ram_gb()
        assert result == 32.0
        mock_macos.assert_called_once()


class TestMacosRAM:
    @patch("subprocess.run")
    def test_parses_sysctl(self, mock_run: MagicMock) -> None:
        mock_run.return_value = subprocess.CompletedProcess(
            args=[],
            returncode=0,
            stdout="17179869184\n",  # 16 GiB
        )
        assert _macos_ram_gb() == pytest.approx(16.0)

    @patch("subprocess.run", side_effect=FileNotFoundError)
    def test_returns_zero_on_failure(self, mock_run: MagicMock) -> None:
        assert _macos_ram_gb() == 0.0
        mock_run.assert_called_once()
